@@ -27,6 +27,7 @@ from hyperliquid.utils.signing import (
     sign_l1_action,
     sign_multi_sig_action,
     sign_spot_transfer_action,
+    sign_token_delegate_action,
     sign_usd_class_transfer_action,
     sign_usd_transfer_action,
     sign_withdraw_from_bridge_action,
@@ -52,6 +53,7 @@ class Exchange(API):
         self.vault_address = vault_address
         self.account_address = account_address
         self.info = Info(base_url, True, meta, spot_meta)
+        self.expires_after: Optional[int] = None
 
     def _post_action(self, action, signature, nonce):
         payload = {
@@ -59,6 +61,7 @@ class Exchange(API):
             "nonce": nonce,
             "signature": signature,
             "vaultAddress": self.vault_address if action["type"] != "usdClassTransfer" else None,
+            "expiresAfter": self.expires_after,
         }
         logging.debug(payload)
         return self.post("/exchange", payload)
@@ -84,6 +87,12 @@ class Exchange(API):
         # We round px to 5 significant figures and 6 decimals for perps, 8 decimals for spot
         return round(float(f"{px:.5g}"), (6 if not is_spot else 8) - self.info.asset_to_sz_decimals[asset])
 
+    # expires_after will cause actions to be rejected after that timestamp in milliseconds
+    # expires_after is not supported on user_signed actions (e.g. usd_transfer) and must be None in order for those
+    # actions to work.
+    def set_expires_after(self, expires_after: Optional[int]) -> None:
+        self.expires_after = expires_after
+
     def order(
         self,
         name: str,
@@ -122,6 +131,7 @@ class Exchange(API):
             order_action,
             self.vault_address,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
 
@@ -176,6 +186,7 @@ class Exchange(API):
             modify_action,
             self.vault_address,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
 
@@ -262,6 +273,7 @@ class Exchange(API):
             cancel_action,
             self.vault_address,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
 
@@ -289,6 +301,7 @@ class Exchange(API):
             cancel_action,
             self.vault_address,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
 
@@ -317,6 +330,7 @@ class Exchange(API):
             schedule_cancel_action,
             self.vault_address,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -338,6 +352,7 @@ class Exchange(API):
             update_leverage_action,
             self.vault_address,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -360,6 +375,7 @@ class Exchange(API):
             update_isolated_margin_action,
             self.vault_address,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -379,6 +395,7 @@ class Exchange(API):
             set_referrer_action,
             None,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -398,6 +415,7 @@ class Exchange(API):
             create_sub_account_action,
             None,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -438,6 +456,7 @@ class Exchange(API):
             sub_account_transfer_action,
             None,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -460,6 +479,7 @@ class Exchange(API):
             sub_account_transfer_action,
             None,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -477,7 +497,7 @@ class Exchange(API):
             "usd": usd,
         }
         is_mainnet = self.base_url == MAINNET_API_URL
-        signature = sign_l1_action(self.wallet, vault_transfer_action, None, timestamp, is_mainnet)
+        signature = sign_l1_action(self.wallet, vault_transfer_action, None, timestamp, self.expires_after, is_mainnet)
         return self._post_action(
             vault_transfer_action,
             signature,
@@ -512,6 +532,23 @@ class Exchange(API):
             timestamp,
         )
 
+    def token_delegate(self, validator: str, wei: int, is_undelegate: bool) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "validator": validator,
+            "wei": wei,
+            "isUndelegate": is_undelegate,
+            "nonce": timestamp,
+            "type": "tokenDelegate",
+        }
+        is_mainnet = self.base_url == MAINNET_API_URL
+        signature = sign_token_delegate_action(self.wallet, action, is_mainnet)
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
     def withdraw_from_bridge(self, amount: float, destination: str) -> Any:
         timestamp = get_timestamp_ms()
         action = {"destination": destination, "amount": str(amount), "time": timestamp, "type": "withdraw3"}
@@ -590,6 +627,7 @@ class Exchange(API):
             action,
             None,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -615,6 +653,75 @@ class Exchange(API):
             action,
             None,
             timestamp,
+            self.expires_after,
+            self.base_url == MAINNET_API_URL,
+        )
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
+    def spot_deploy_enable_freeze_privilege(self, token: int) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "type": "spotDeploy",
+            "enableFreezePrivilege": {
+                "token": token,
+            },
+        }
+        signature = sign_l1_action(
+            self.wallet,
+            action,
+            None,
+            timestamp,
+            self.expires_after,
+            self.base_url == MAINNET_API_URL,
+        )
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
+    def spot_deploy_freeze_user(self, token: int, user: str, freeze: bool) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "type": "spotDeploy",
+            "freezeUser": {
+                "token": token,
+                "user": user.lower(),
+                "freeze": freeze,
+            },
+        }
+        signature = sign_l1_action(
+            self.wallet,
+            action,
+            None,
+            timestamp,
+            self.expires_after,
+            self.base_url == MAINNET_API_URL,
+        )
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
+    def spot_deploy_revoke_freeze_privilege(self, token: int) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "type": "spotDeploy",
+            "revokeFreezePrivilege": {
+                "token": token,
+            },
+        }
+        signature = sign_l1_action(
+            self.wallet,
+            action,
+            None,
+            timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -640,6 +747,7 @@ class Exchange(API):
             action,
             None,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -661,6 +769,7 @@ class Exchange(API):
             action,
             None,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -690,6 +799,154 @@ class Exchange(API):
             action,
             None,
             timestamp,
+            self.expires_after,
+            self.base_url == MAINNET_API_URL,
+        )
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
+    def spot_deploy_set_deployer_trading_fee_share(self, token: int, share: str) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "type": "spotDeploy",
+            "setDeployerTradingFeeShare": {
+                "token": token,
+                "share": share,
+            },
+        }
+        signature = sign_l1_action(
+            self.wallet,
+            action,
+            None,
+            timestamp,
+            self.expires_after,
+            self.base_url == MAINNET_API_URL,
+        )
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
+    def c_signer_unjail_self(self) -> Any:
+        return self.c_signer_inner("unjailSelf")
+
+    def c_signer_jail_self(self) -> Any:
+        return self.c_signer_inner("jailSelf")
+
+    def c_signer_inner(self, variant: str) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "type": "CSignerAction",
+            variant: None,
+        }
+        signature = sign_l1_action(
+            self.wallet,
+            action,
+            None,
+            timestamp,
+            self.expires_after,
+            self.base_url == MAINNET_API_URL,
+        )
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
+    def c_validator_register(
+        self,
+        node_ip: str,
+        name: str,
+        description: str,
+        delegations_disabled: bool,
+        commission_bps: int,
+        signer: str,
+        unjailed: bool,
+        initial_wei: int,
+    ) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "type": "CValidatorAction",
+            "register": {
+                "profile": {
+                    "node_ip": {"Ip": node_ip},
+                    "name": name,
+                    "description": description,
+                    "delegations_disabled": delegations_disabled,
+                    "commission_bps": commission_bps,
+                    "signer": signer,
+                },
+                "unjailed": unjailed,
+                "initial_wei": initial_wei,
+            },
+        }
+        signature = sign_l1_action(
+            self.wallet,
+            action,
+            None,
+            timestamp,
+            self.expires_after,
+            self.base_url == MAINNET_API_URL,
+        )
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
+    def c_validator_change_profile(
+        self,
+        node_ip: Optional[str],
+        name: Optional[str],
+        description: Optional[str],
+        unjailed: bool,
+        disable_delegations: Optional[bool],
+        commission_bps: Optional[int],
+        signer: Optional[str],
+    ) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "type": "CValidatorAction",
+            "changeProfile": {
+                "node_ip": None if node_ip is None else {"Ip": node_ip},
+                "name": name,
+                "description": description,
+                "unjailed": unjailed,
+                "disable_delegations": disable_delegations,
+                "commission_bps": commission_bps,
+                "signer": signer,
+            },
+        }
+        signature = sign_l1_action(
+            self.wallet,
+            action,
+            None,
+            timestamp,
+            self.expires_after,
+            self.base_url == MAINNET_API_URL,
+        )
+        return self._post_action(
+            action,
+            signature,
+            timestamp,
+        )
+
+    def c_validator_unregister(self) -> Any:
+        timestamp = get_timestamp_ms()
+        action = {
+            "type": "CValidatorAction",
+            "unregister": None,
+        }
+        signature = sign_l1_action(
+            self.wallet,
+            action,
+            None,
+            timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
@@ -717,6 +974,7 @@ class Exchange(API):
             is_mainnet,
             vault_address,
             nonce,
+            self.expires_after,
         )
         return self._post_action(
             multi_sig_action,
@@ -735,6 +993,7 @@ class Exchange(API):
             action,
             None,
             timestamp,
+            self.expires_after,
             self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
